#include "qt_ble.h"

#define MSEC_TO_UNITS(TIME, RESOLUTION)                         (((TIME) * 1000) / (RESOLUTION))

#define NUS_BASE_UUID                                           { { 0x59, 0x46, 0x2F, 0x12, 0x95, 0x43, 0x99, 0x99, 0x12, 0xC8, 0x58, 0xB4, 0x59, 0xA2, 0x71, 0x2D } }
#define BLE_UUID_DIS_SERVICE                                    0x0001
#define BLE_UUID_NUS_RX_CHARACTERISTIC                          0x0002
#define BLE_UUID_NUS_TX_CHARACTERISTIC                          0x0003
#define BLE_UUID_CCCD                                           0x2902

#define BLE_CCCD_NOTIFY                                         0x01

static QString _mDeviceName;

static uint32_t _mErrorCode             = 0;
static uint32_t _mConfigID              = 0;
static uint16_t _mMTU                   = 0;
static uint16_t _mDeviceServiceUUID     = 0;
static uint16_t _mCCCDHandle            = 0;
static uint16_t _mRxHandle              = 0;
static uint16_t _mTxHandle              = 0;
static uint16_t _mCharHandle            = 0;
static uint16_t _mConnectionHandle      = 0;
static uint16_t _mServiceStartHandle    = 0;
static uint16_t _mServiceEndHandle      = 0;

static bool _mIsConnected               = false;

static ble_uuid_t UUID;
static ble_cfg_t _mBLE_Config;
static adapter_t* _mAdapter;
static physical_layer_t* _mPhy;
static data_link_layer_t* _mData_link_layer;
static transport_layer_t* _mTransport_layer;

static const ble_gap_scan_params_t Scan_Params =
{
    0,                       // Set if active scanning.
    0,                       // Set if selective scanning.
    0,
    (uint16_t)0x00A0,
    (uint16_t)0x0050,
    (uint16_t)0x00
};

static const ble_gap_conn_params_t Connection_Params =
{
    (uint16_t)MSEC_TO_UNITS(7.5, 1250),
    (uint16_t)MSEC_TO_UNITS(7.5, 1250),
    (uint16_t)0,
    (uint16_t)MSEC_TO_UNITS(4000, 10000)
};

uint32_t Qt_BLE::Open(QString Port, QString DeviceName, uint32_t ServiceUUID, uint16_t MTU)
{
    const uint32_t RAM_Start = 0;
    uint32_t* App_RAM_Base = NULL;

    qRegisterMetaType<uint16_t>("uint16_t");

    _mMTU = MTU;
    _mDeviceName = DeviceName;
    _mDeviceServiceUUID = ServiceUUID;
    _mConfigID = 1;
    _mConnectionHandle = 0;
    _mIsConnected = false;

    LogSend("Open softdevice...");
    _mPhy = sd_rpc_physical_layer_create_uart(Port.toLocal8Bit().data(), 1000000, SD_RPC_FLOW_CONTROL_NONE, SD_RPC_PARITY_NONE);
    _mData_link_layer = sd_rpc_data_link_layer_create_bt_three_wire(_mPhy, 250);
    _mTransport_layer = sd_rpc_transport_layer_create(_mData_link_layer, 1500);
    _mAdapter = sd_rpc_adapter_create(_mTransport_layer);

    _mErrorCode = sd_rpc_log_handler_severity_filter_set(_mAdapter, SD_RPC_LOG_INFO) | sd_rpc_open(_mAdapter, StatusHandler, Qt_BLE::on_Event, Qt_BLE::LogHandler);
    if(_mErrorCode != NRF_SUCCESS)
    {
        LogSend(QString("Failed to open nRF BLE Driver. Error code: 0x%1").arg(_mErrorCode, 4, 16, QLatin1Char('0')));
    }

    _mBLE_Config.gap_cfg.role_count_cfg.periph_role_count    = 0;
    _mBLE_Config.gap_cfg.role_count_cfg.central_role_count   = 1;
    _mBLE_Config.gap_cfg.role_count_cfg.central_sec_count    = 0;

    _mErrorCode = sd_ble_cfg_set(_mAdapter, BLE_GAP_CFG_ROLE_COUNT, &_mBLE_Config, RAM_Start);
    if(_mErrorCode != NRF_SUCCESS)
    {
        LogSend(QString("sd_ble_cfg_set failed when attempting to set BLE_GAP_CFG_ROLE_COUNT. Error code: 0x%1").arg(_mErrorCode, 4, 16, QLatin1Char('0')));
    }

    memset(&_mBLE_Config, 0x00, sizeof(ble_cfg_t));
    _mBLE_Config.conn_cfg.conn_cfg_tag                        = _mConfigID;
    _mBLE_Config.conn_cfg.params.gatt_conn_cfg.att_mtu        = 150;

    _mErrorCode = sd_ble_cfg_set(_mAdapter, BLE_CONN_CFG_GATT, &_mBLE_Config, RAM_Start);
    if(_mErrorCode != NRF_SUCCESS)
    {
        LogSend(QString("sd_ble_cfg_set failed when attempting to set BLE_CONN_CFG_GATT. Error code: 0x%1").arg(_mErrorCode, 4, 16, QLatin1Char('0')));
    }

    LogSend("Enable softdevice...");
    _mErrorCode = sd_ble_enable(_mAdapter, App_RAM_Base);
    if(_mErrorCode != NRF_SUCCESS)
    {
        LogSend(QString("Failed to enable nRF BLE Driver. Error code: 0x%1").arg(_mErrorCode, 4, 16, QLatin1Char('0')));
    }

    return _mErrorCode;
}

uint32_t Qt_BLE::StartScan(void)
{
    LogSend("Start scanning...");

    _mErrorCode = sd_ble_gap_scan_start(_mAdapter, &Scan_Params);
    if(_mErrorCode != NRF_SUCCESS)
    {
        LogSend(QString("Scan start failed. Error code: 0x%1").arg(_mErrorCode, 4, 16, QLatin1Char('0')));
    }

    return _mErrorCode;
}

uint32_t Qt_BLE::Close(void)
{
    if(_mIsConnected)
    {
        _mErrorCode = sd_ble_gap_disconnect(_mAdapter, _mConnectionHandle, 0x13) | sd_rpc_close(_mAdapter);
        if(_mErrorCode != NRF_SUCCESS)
        {
            return _mErrorCode;
        }

        _mIsConnected = false;
    }

    return NRF_SUCCESS;
}

uint32_t Qt_BLE::Transmit(uint8_t* p_Data, uint16_t Length, uint16_t Offset)
{
    ble_gattc_write_params_t WriteParams;

    if(_mRxHandle == 0)
    {
        LogSend("No Rx handle has been found!");

        return NRF_ERROR_INVALID_STATE;
    }

    WriteParams.handle     = _mRxHandle;
    WriteParams.len        = Length;
    WriteParams.p_value    = p_Data;
    WriteParams.write_op   = BLE_GATT_OP_WRITE_REQ;
    WriteParams.offset     = Offset;

    return sd_ble_gattc_write(_mAdapter, _mConnectionHandle, &WriteParams);
}

bool Qt_BLE::isOpen(void)
{
    return _mIsConnected;
}

Qt_BLE& Qt_BLE::Instance(void)
{
    static Qt_BLE _mInstance;

    return _mInstance;
}

void Qt_BLE::LogSend(QString Message)
{
    emit Qt_BLE::Instance().on_LogSend(Message);
}

uint32_t Qt_BLE::Characteristics_Discovery_Start(void)
{
    ble_gattc_handle_range_t HandleRange;

    LogSend( "Discovering characteristics");

    _mServiceStartHandle++;

    HandleRange.start_handle = _mServiceStartHandle;
    HandleRange.end_handle = _mServiceEndHandle;

    return sd_ble_gattc_characteristics_discover(_mAdapter, _mConnectionHandle, &HandleRange);
}

uint32_t Qt_BLE::Descriptor_Discovery_Start(void)
{
    ble_gattc_handle_range_t HandleRange;

    LogSend("Discovering descriptors");

    _mServiceStartHandle++;

    HandleRange.start_handle   = _mServiceStartHandle;
    HandleRange.end_handle     = _mServiceEndHandle;

    return sd_ble_gattc_descriptors_discover(_mAdapter, _mConnectionHandle, &HandleRange);
}

void Qt_BLE::StatusHandler(adapter_t*, sd_rpc_app_status_t Code, const char* p_Message)
{
    emit Qt_BLE::Instance().on_Status(Code, QString(p_Message));
}

void Qt_BLE::LogHandler(adapter_t*, sd_rpc_log_severity_t, const char* p_Message)
{
    LogSend(QString("%1").arg(p_Message));
}

void Qt_BLE::AddressToString(ble_gap_addr_t Address, rsize_t Size, uint8_t* p_Buffer)
{
    char Temp[3];

    for(int i = 5; i >= 0; --i)
    {
        sprintf(Temp, "%02X", Address.addr[i]);
        strcat_s((char*)p_Buffer, Size, Temp);
    }
}

bool Qt_BLE::Find_Adv_Name(const ble_gap_evt_adv_report_t* p_Adv_Report, const char* p_Name)
{
    data_t DeviceName;
    data_t AdvertisingData;

    AdvertisingData.p_Data = (uint8_t*)p_Adv_Report->data;
    AdvertisingData.Length = p_Adv_Report->dlen;

    // Search for advertising names
    _mErrorCode = Parse_Adv_Report(BLE_GAP_AD_TYPE_COMPLETE_LOCAL_NAME, &AdvertisingData, &DeviceName);
    if(_mErrorCode == NRF_SUCCESS)
    {
        if(memcmp(p_Name, DeviceName.p_Data, DeviceName.Length) == 0)
        {
            return true;
        }
    }
    else
    {
        // Look for the short local name if it was not found as complete
        _mErrorCode = Parse_Adv_Report(BLE_GAP_AD_TYPE_SHORT_LOCAL_NAME,  &AdvertisingData, &DeviceName);
        if(_mErrorCode != NRF_SUCCESS)
        {
            return false;
        }

        if(memcmp(p_Name, DeviceName.p_Data, DeviceName.Length) == 0)
        {
            return true;
        }
    }

    return false;
}

uint32_t Qt_BLE::Parse_Adv_Report(uint8_t type, data_t* p_AdvData, data_t* p_TypeData)
{
    uint32_t Index = 0;
    uint8_t* p_Data;

    p_Data = p_AdvData->p_Data;

    while(Index < p_AdvData->Length)
    {
        uint8_t field_length = p_Data[Index];
        uint8_t field_type   = p_Data[Index + 1];

        if(field_type == type)
        {
            p_TypeData->p_Data = &p_Data[Index + 2];
            p_TypeData->Length = field_length - 1;

            return NRF_SUCCESS;
        }

        Index += field_length + 1;
    }

    return NRF_ERROR_NOT_FOUND;
}

uint32_t Qt_BLE::Start_ServiceDiscovery(adapter_t* p_Adapter, uint32_t Start)
{
    LogSend("Discovering primary services");

    /*
    _mErrorCode = sd_ble_uuid_vs_add(p_Adapter, &Base_UUID, &UUID.type);
    if(_mErrorCode != NRF_SUCCESS)
    {
        LogSend(QString("Failed to get UUID. Error code: 0x%1").arg(_mErrorCode, 4, 16, QLatin1Char('0')));
    }

    UUID.uuid = _mDeviceServiceUUID;
*/
    _mErrorCode = sd_ble_gattc_primary_services_discover(p_Adapter, _mConnectionHandle, Start, NULL);
    if(_mErrorCode != NRF_SUCCESS)
    {
        LogSend(QString("Failed to initiate or continue a GATT Primary Service Discovery procedure. Error code: 0x%1").arg(_mErrorCode, 4, 16, QLatin1Char('0')));
    }

    return _mErrorCode;
}

void Qt_BLE::on_Adv_Report(adapter_t* p_Adapter, const ble_gap_evt_t* const p_Event)
{
    uint8_t Buffer[50] = {0};

    AddressToString(p_Event->params.adv_report.peer_addr, sizeof(Buffer), Buffer);

    if(Find_Adv_Name(&p_Event->params.adv_report, _mDeviceName.toLocal8Bit()))
    {
        _mErrorCode = sd_ble_gap_connect(p_Adapter, &(p_Event->params.adv_report.peer_addr), &Scan_Params, &Connection_Params, _mConfigID);
        if(_mErrorCode != NRF_SUCCESS)
        {
            LogSend(QString("Connection Request Failed. Error code: 0x%1").arg(_mErrorCode, 4, 16, QLatin1Char('0')));
            return;
        }
    }
}

void Qt_BLE::on_Connected(adapter_t* p_Adapter, const ble_gap_evt_t* const p_Event)
{
    LogSend("Connection established");

    _mConnectionHandle = p_Event->conn_handle;
    _mIsConnected = true;

    Start_ServiceDiscovery(p_Adapter);
}

void Qt_BLE::on_Timeout(const ble_gap_evt_t* const p_Event)
{
    if(p_Event->params.timeout.src == BLE_GAP_TIMEOUT_SRC_CONN)
    {
        _mIsConnected = false;
    }
    else if(p_Event->params.timeout.src == BLE_GAP_TIMEOUT_SRC_SCAN)
    {
        StartScan();
    }

    emit Qt_BLE::Instance().on_Timeout();
}

void Qt_BLE::on_Service_Discovery_Response(const ble_gattc_evt_t* const p_Event)
{
    uint32_t ServiceIndex = 0;
    const ble_gattc_service_t* Service;

    LogSend(QString("Received service discovery response. Service count: %1").arg(p_Event->params.prim_srvc_disc_rsp.count));

    if(p_Event->gatt_status != NRF_SUCCESS)
    {
        LogSend(QString("       Service discovery failed. Error code: 0x%1").arg(p_Event->gatt_status, 4, 16, QLatin1Char('0')));

        return;
    }
    else if(p_Event->params.prim_srvc_disc_rsp.count == 0)
    {
        LogSend("       Service not found");

        return;
    }

    for(uint32_t i = 0; i < p_Event->params.prim_srvc_disc_rsp.count; i++)
    {
        Service = &(p_Event->params.prim_srvc_disc_rsp.services[i]);

        LogSend(QString("  Service UUID: 0x%1").arg(Service->uuid.uuid, 4, 16, QLatin1Char('0')));
    }

    if(Service->uuid.uuid != 0x180A)
    {
        LogSend(QString("  Unknown service discovered with UUID: 0x%1").arg(Service->uuid.uuid, 4, 16, QLatin1Char('0')));

        return;
    }

    _mServiceStartHandle = Service->handle_range.start_handle;
    _mServiceEndHandle = Service->handle_range.end_handle;

    LogSend(QString("   Discovered service with UUID: 0x%1, start handle: 0x%2, end handle: 0x%3").arg(Service->uuid.uuid, 4, 16, QLatin1Char('0')).arg(_mServiceEndHandle, 4, 16, QLatin1Char('0')).arg(_mServiceEndHandle, 4, 16, QLatin1Char('0')));

    Characteristics_Discovery_Start();
}

void Qt_BLE::on_Characteristic_Discovery_Response(const ble_gattc_evt_t* const p_Event)
{
    LogSend(QString("Received characteristic discovery response. Characteristics count: %1").arg(p_Event->params.char_disc_rsp.count));

    if(p_Event->gatt_status != NRF_SUCCESS)
    {
        LogSend(QString("  Characteristic discovery failed. Error code: 0x%1").arg(p_Event->gatt_status, 4, 16, QLatin1Char('0')));

        return;
    }
    else if(p_Event->params.char_disc_rsp.count == 0)
    {
        LogSend("  No characteristic handle found");

        return;
    }

    for(int i = 0; i < p_Event->params.char_disc_rsp.count; i++)
    {
        LogSend(QString("  Characteristic handle: 0x%1, UUID: 0x%2").arg(p_Event->params.char_disc_rsp.chars[i].handle_decl, 4, 16, QLatin1Char('0')).arg(p_Event->params.char_disc_rsp.chars[i].uuid.uuid, 4, 16, QLatin1Char('0')));

        if(p_Event->params.char_disc_rsp.chars[i].uuid.uuid == BLE_UUID_NUS_RX_CHARACTERISTIC)
        {
            _mCharHandle = p_Event->params.char_disc_rsp.chars[i].handle_decl;
            Descriptor_Discovery_Start();
        }
        else if(p_Event->params.desc_disc_rsp.descs[i].uuid.uuid == BLE_UUID_NUS_TX_CHARACTERISTIC)
        {
            _mTxHandle =  p_Event->params.char_disc_rsp.chars[i].handle_decl;
            Descriptor_Discovery_Start();
        }
    }
}

void Qt_BLE::on_Descriptor_Discovery_Response(const ble_gattc_evt_t* const p_Event)
{
    if(p_Event->gatt_status != NRF_SUCCESS)
    {
        LogSend(QString("Descriptor discovery failed. Error code: 0x%1").arg(p_Event->gatt_status, 4, 16, QLatin1Char('0')));

        return;
    }

    LogSend(QString("Received descriptor discovery response. Descriptor count: %1").arg(p_Event->params.desc_disc_rsp.count));

    for(int i = 0; i < p_Event->params.desc_disc_rsp.count; i++)
    {
        LogSend(QString("   Descriptor handle: 0x%1, UUID: 0x%2").arg(p_Event->params.desc_disc_rsp.descs[i].handle, 4, 16, QLatin1Char('0')).arg(p_Event->params.desc_disc_rsp.descs[i].uuid.uuid, 4, 16, QLatin1Char('0')));

        if(p_Event->params.desc_disc_rsp.descs[i].uuid.uuid == BLE_UUID_CCCD)
        {
            ble_gattc_write_params_t WriteParams;
            uint8_t                  CCCD[2] = {BLE_CCCD_NOTIFY, 0x00};

            _mCCCDHandle = p_Event->params.desc_disc_rsp.descs[i].handle;

            LogSend("       Setting CCCD");

            if(_mCCCDHandle == 0)
            {
                LogSend("       No CCCD handle has been found");

                return;
            }

            WriteParams.handle     = _mCCCDHandle;
            WriteParams.len        = sizeof(CCCD) / sizeof(uint8_t);
            WriteParams.p_value    = CCCD;
            WriteParams.write_op   = BLE_GATT_OP_WRITE_REQ;
            WriteParams.offset     = 0;

            sd_ble_gattc_write(_mAdapter, _mConnectionHandle, &WriteParams);
        }
        else if(p_Event->params.desc_disc_rsp.descs[i].uuid.uuid == BLE_UUID_NUS_RX_CHARACTERISTIC)
        {
            _mRxHandle = p_Event->params.desc_disc_rsp.descs[i].handle;
        }
        else if(p_Event->params.desc_disc_rsp.descs[i].uuid.uuid == BLE_UUID_NUS_TX_CHARACTERISTIC)
        {
            _mTxHandle = p_Event->params.desc_disc_rsp.descs[i].handle;
        }
    }

    Descriptor_Discovery_Start();
}

void Qt_BLE::on_Write_Response(const ble_gattc_evt_t* const p_Event)
{
    LogSend("Received write response.");

    if(p_Event->gatt_status != NRF_SUCCESS)
    {
        LogSend(QString("Error. Write operation failed. Error code: 0x%1").arg(p_Event->gatt_status, 4, 16, QLatin1Char('0')));
    }
}

void Qt_BLE::on_Conn_Params_Update_Request(const ble_gap_evt_t* const p_Event)
{
    _mErrorCode = sd_ble_gap_conn_param_update(_mAdapter, _mConnectionHandle, &(p_Event->params.conn_param_update_request.conn_params));
    if(_mErrorCode != NRF_SUCCESS)
    {
        LogSend(QString("Conn params update failed. Error code: 0x%1").arg(_mErrorCode, 4, 16, QLatin1Char('0')));
    }
}

void Qt_BLE::on_Exchange_MTU_Request(const ble_gatts_evt_t* const)
{
    _mErrorCode = sd_ble_gatts_exchange_mtu_reply(_mAdapter, _mConnectionHandle, _mMTU);
    if(_mErrorCode != NRF_SUCCESS)
    {
        LogSend(QString("MTU exchange request reply failed. Error code: 0x%1").arg(_mErrorCode, 4, 16, QLatin1Char('0')));
    }
}

void Qt_BLE::on_Conn_Data_Length_Update_Request(const ble_gattc_evt_t* const p_Event)
{
    LogSend("BLE_GAP_EVT_DATA_LENGTH_UPDATE_REQUEST received");

    ble_gap_data_length_limitation_t dll = { 0 };
    ble_gap_data_length_params_t dlp;

    dlp.max_tx_octets = 27;
    dlp.max_rx_octets = 27;
    dlp.max_tx_time_us = BLE_GAP_DATA_LENGTH_AUTO;
    dlp.max_rx_time_us = BLE_GAP_DATA_LENGTH_AUTO;

    _mErrorCode = sd_ble_gap_data_length_update(_mAdapter, _mConnectionHandle, &dlp, &dll);
    if(_mErrorCode != NRF_SUCCESS)
    {
        LogSend(QString("sd_ble_gap_data_length_update failed. Error code: 0x%1").arg(_mErrorCode, 4, 16, QLatin1Char('0')));
    }
}

void Qt_BLE::on_HVX(const ble_gattc_evt_t* const p_Event)
{
    if((p_Event->params.hvx.handle >= _mTxHandle) || (p_Event->params.hvx.handle <= _mCCCDHandle))
    {
        emit Qt_BLE::Instance().on_ReceiveComplete(p_Event->params.hvx.data, p_Event->params.hvx.len);
    }
    else
    {
        LogSend(QString("Unparsed data received on handle: 0x%1").arg(p_Event->params.hvx.handle, 4, 16, QLatin1Char('0')));
    }
}

void Qt_BLE::on_Exchange_MTU_Response(const ble_gattc_evt_t* const p_Event)
{
    uint16_t server_rx_mtu = p_Event->params.exchange_mtu_rsp.server_rx_mtu;

    LogSend(QString("MTU response received. New ATT_MTU is %1").arg(server_rx_mtu));
}

void Qt_BLE::on_Event(adapter_t* p_Adapter, ble_evt_t* p_Event)
{
    LogSend(QString("ID: 0x%1").arg(p_Event->header.evt_id, 4, 16, QLatin1Char('0')));

    switch(p_Event->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
        {
            LogSend("Connected");

            Qt_BLE::on_Connected(p_Adapter, &(p_Event->evt.gap_evt));

            break;
        }
        case BLE_GAP_EVT_DISCONNECTED:
        {
            LogSend("Disconnected");

            LogSend(QString("Disconnected, reason: 0x%1").arg(p_Event->evt.gap_evt.params.disconnected.reason, 4, 16, QLatin1Char('0')));

            _mConnectionHandle = 0;

            emit Qt_BLE::Instance().on_Disconnect();

            break;
        }
        case BLE_GAP_EVT_TIMEOUT:
        {
            LogSend("Timeout");

            Qt_BLE::on_Timeout(&(p_Event->evt.gap_evt));

            break;
        }
        case BLE_GAP_EVT_ADV_REPORT:
        {
            Qt_BLE::on_Adv_Report(p_Adapter, &(p_Event->evt.gap_evt));

            break;
        }
        case BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST:
        {
             Qt_BLE::on_Conn_Params_Update_Request(&(p_Event->evt.gap_evt));

             break;
        }
        case BLE_GAP_EVT_CONN_PARAM_UPDATE:
        {
            LogSend("Connection paramaters updated");
            LogSend(QString("   Connection interval: %1").arg(p_Event->evt.gap_evt.params.conn_param_update.conn_params.max_conn_interval));
            LogSend(QString("   Supervision timeout: %1").arg(p_Event->evt.gap_evt.params.conn_param_update.conn_params.conn_sup_timeout / 100));
            LogSend(QString("   Slave latency: %1").arg(p_Event->evt.gap_evt.params.conn_param_update.conn_params.slave_latency));

            break;
        }
        case BLE_GAP_EVT_DATA_LENGTH_UPDATE_REQUEST:
        {
            Qt_BLE::on_Conn_Data_Length_Update_Request(&(p_Event->evt.gattc_evt));

            break;
        }
        case BLE_GATTC_EVT_PRIM_SRVC_DISC_RSP:
        {
            Qt_BLE::on_Service_Discovery_Response(&(p_Event->evt.gattc_evt));

            break;
        }
        case BLE_GATTC_EVT_CHAR_DISC_RSP:
        {
            Qt_BLE::on_Characteristic_Discovery_Response(&(p_Event->evt.gattc_evt));

            break;
        }
        case BLE_GATTC_EVT_DESC_DISC_RSP:
        {
            Qt_BLE::on_Descriptor_Discovery_Response(&(p_Event->evt.gattc_evt));

            break;
        }
        case BLE_GATTC_EVT_WRITE_RSP:
        {
            Qt_BLE::on_Write_Response(&(p_Event->evt.gattc_evt));

            break;
        }
        case BLE_GATTC_EVT_HVX:
        {
            Qt_BLE::on_HVX(&(p_Event->evt.gattc_evt));

            break;
        }
        case BLE_GATTS_EVT_EXCHANGE_MTU_REQUEST:
        {
            Qt_BLE::on_Exchange_MTU_Request(&(p_Event->evt.gatts_evt));

            break;
        }
        case BLE_GATTC_EVT_EXCHANGE_MTU_RSP:
        {
            Qt_BLE::on_Exchange_MTU_Response(&(p_Event->evt.gattc_evt));

            break;
        }
        case BLE_GATTC_EVT_WRITE_CMD_TX_COMPLETE:
        {
            emit Qt_BLE::Instance().on_TransmissionComplete();

            break;
        }
        default:
        {
            LogSend(QString("Received an unhandled event with ID: 0x%1").arg(p_Event->header.evt_id, 4, 16, QLatin1Char('0')));

            break;
        }
    }
}
